/*
** Module   : AdminACL.java
** Abstract : Access Control List management client
**
** Copyright (c) 2009-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 CA  20090714  @43202  Created initial version.
** 002 CA  20090720  @43240  Added handlers for the ACL actions (simple and 
**                           complex).
** 003 NVS 20090723  @43349  Massive update to most actions. Added support for
**                           ACLs with errors throughout the file.
** 004 NVS 20090723  @43380  Reworked ACL candidate ID calculation.
** 005 NVS 20090724  @43389  Add ACL dialog no longer depends on the table
**                           selection. ACL instance and resource type can be
**                           selected from the combo boxes. However, selection
**                           may provide better defaults.
** 006 NVS 20090727  @43397  Add ACL dialog has subjects and rights editing
**                           capabilities and is complete.
** 007 NVS 20090729  @43425  Implemented ACL printing.
** 008 NVS 20090803  @43478  Using a better approach to printing with extended
**                           detail choice.
** 009 NVS 20090804  @43511  Fixed bug in subjects editor doing wrong index
**                           conversion.
** 010 NVS 20090810  @43583  Fixed bug in show by resource getting empty set.
** 011 NVS 20090930  @44101  Extended the edit() method to take the resource 
**                           instance name associated with this instance of 
**                           rights.
** 012 NVS 20091116  @44344  Eliminating confusing empty Admin Client messages.
** 013 SIY 20091116  @44337  Refactoring of things related to TaggedName. 
**                           Minor cleanups.
** 014 SIY 20091204  @44480  Modified for use in standalone mode in function 
**                           maintenance.
** 015 CS  20121116          identify and remove any customer-specific 
**                           references inside the P2J project: moved client
**                           ACL reference to the client implementation of  
*                            ClientAccountExtension which calls this class's 
*                            public showByResourceType() method.
 */

package com.goldencode.p2j.admin.client;

import java.util.*;
import java.util.List;
import java.awt.*;
import java.awt.event.*;
import java.awt.print.PrinterException;
import java.text.MessageFormat;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.*;

import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.security.*;

/**
 * This is a Access Control List management class for the admin session. 
 * 
 * @author CA
 */
public class AdminACL
extends AdminScreen
implements PrintRequest
{
   /** Screen switcher - ACL screen ID */
   static final String SCREEN_ACLS = "adm_acls";
   
   /** Screen title */
   static final String TITLE_ACLS = "Access Control";
   
   /** Shared instance name */
   private static final String SHARED_INSTANCE = "*shared*";
   
   // center panel widgets, all named cp*
   
   /** center panel - works with the list of ACLs */
   private JPanel center = null;
   
   /** Effective View check-box */
   private JCheckBox cpcEv = null;

   /** scroller for list of ACLs */
   private JScrollPane cpScroller = null;
   
   /** table to display all ACLs */
   private JTable cpACLs = null;

   /** related table model */
   private AclTableModel cpACLtm = null;
   
   /** related table sorter */
   private TableRowSorter<AclTableModel> cpACLsorter = null;

   /** related table selection model */
   private ListSelectionModel cpACLsm = null;

   /** sub-panel that holds all controls for list of ACLs */
   private JPanel cpCtlPanel = null;
   
   /** Populate button */
   private JButton cpbR = null;

   /** Add ACL button */
   private JButton cpbA = null;
   
   /** Delete Selected button */
   private JButton cpbD = null;

   /** Clone ACL button */
   private JButton cpbC = null;
   
   /** Move ACL Up button */
   private JButton cpbMu = null;
   
   /** Move ACL Down button */
   private JButton cpbMd = null;
   
   /** Filter ACLs button */
   private JButton cpbF = null;
   
   /** Resequence ACLs button */
   private JButton cpbS = null;

   /** Cancel button in standalone mode.  */
   private JButton cpbCl = null;
   
   // dialog for resequence ACLs specific widgets, all named cpsd*

   /** Start id label */
   private JLabel cpsdlId = null;
   
   /** Start id field */
   private JTextField cpsdfId = null;
   
   /** Step label */
   private JLabel cpsdlStep = null;

   /** Step field */
   private JTextField cpsdfStep = null;
  
   /** Resequence button in modal dialog*/
   private JButton cpsdbRe = null;
   
   /** Close button in Resequence modal dialog*/
   private JButton cpsdbC = null;
   
   /** modal dialog panel for Resequence ACLs function */
   private JPanel cpsPanel = null;
   
   // dialog for ACL resource instance name specific widgets, all named cprd*
   
   /** Resource Instance Name label */
   private JLabel cprdlName = null;
   
   /** Resource Instance Name field */
   private JTextField cprdfName = null;
   
   /** Resource Instance Mode exact label */
   private JLabel cprdlMode1 = null;
   
   /** Resource Instance Mode exact check-box */
   private JRadioButton cprdchkMode1 = null;
   
   /** Resource Instance Mode regexp label */
   private JLabel cprdlMode2 = null;
   
   /** Resource Instance Mode regexp check-box */
   private JRadioButton cprdchkMode2 = null;
   
   /** Save button in modal dialog*/
   private JButton cprdbS = null;
   
   /** Close button in modal dialog*/
   private JButton cprdbC = null;
   
   /** modal dialog panel for change resource instance name function */
   private JPanel cprPanel = null;
   
   // dialog for clone ACL specific widgets, all named cpcd*
   
   /** Clone ID label */
   private JLabel cpcdlId = null;
   
   /** Clone ID field */
   private JTextField cpcdfId = null;
   
   /** Save button in modal dialog*/
   private JButton cpcdbS = null;
   
   /** Close button in modal dialog*/
   private JButton cpcdbC = null;
   
   /** modal dialog panel for clone ACL function */
   private JPanel cpcPanel = null;
   
   // dialog for ACL instance specific widgets, all named cpid*
   
   /** Instance label */
   private JLabel cpidlI = null;
   
   /** Instance field */
   private JTextField cpidfI = null;
   
   /** Shared instance label */
   private JLabel cpidlSh = null;
   
   /** Shared instance check-box */
   private JCheckBox cpidchkSh = null;
   
   /** Save button in modal dialog*/
   private JButton cpidbS = null;
   
   /** Close button in modal dialog*/
   private JButton cpidbC = null;
   
   /** modal dialog panel for change instance function */
   private JPanel cpiPanel = null;
   
   // dialog for direct ACL ID editing, all named cpcd*
   
   /** Target ID label */
   private JLabel cpedlId = null;
   
   /** Target ID field */
   private JTextField cpedfId = null;
   
   /** Save button in modal dialog*/
   private JButton cpedbS = null;
   
   /** Cancel button in modal dialog*/
   private JButton cpedbC = null;
   
   /** modal dialog panel for ID editing */
   private JPanel cpePanel = null;
   
   // dialog for Add ACL Item specific widgets, all named cpdd*
   
   /** ACL Instance field label */
   private JLabel cpddlInst = null;
   
   /** ACL Instance combo box */
   private JComboBox cpddxInst = null;
   
   /** Resource Type field label */
   private JLabel cpddlRtype = null;
   
   /** Resource Type combo box */
   private JComboBox cpddxRtype = null;
   
   /** Item ID label */
   private JLabel cpddlId = null;
   
   /** Item ID field */
   private JTextField cpddfId = null;
   
   /** Resource Instance Name label */
   private JLabel cpddlName = null;
   
   /** Resource Instance Name field */
   private JTextField cpddfName = null;
   
   /** Resource Instance Mode exact label */
   private JLabel cpddlMode1 = null;
   
   /** Resource Instance Mode exact radio button */
   private JRadioButton cpddchkMode1 = null;
   
   /** Resource Instance Mode regexp label */
   private JLabel cpddlMode2 = null;
   
   /** Resource Instance Mode regexp radio button */
   private JRadioButton cpddchkMode2 = null;
   
   /** Subjects field label */
   private JLabel cpddlSubjs = null;
   
   /** Subjects field */
   private JTextField cpddfSubjs = null;
   
   /** Rights field label */
   private JLabel cpddlRights = null;
   
   /** Rights field */
   private JTextField cpddfRights = null;
   
   /** Save button in modal dialog*/
   private JButton cpddbS = null;
   
   /** Close button in modal dialog*/
   private JButton cpddbC = null;
   
   /** modal dialog panel for Add ACL item function */
   private JPanel cpddPanel = null;
   
   /** data model for the ACL instances combo box */
   private DefaultComboBoxModel cpddModI = null;

   /** data model for the resource types combo box */
   private DefaultComboBoxModel cpddModR = null;
   
   // dialog for ACL subjects specific widgets, all named cpid*
   
   /** scroller for list of available accounts */
   private JScrollPane cpadScroller1 = null;
   
   /** table to display all available accounts */
   private JTable cpadAccounts1 = null;

   /** related table model */
   private TaggedDataTableModel cpadAccounttm1 = null;
   
   /** related table selection model */
   private ListSelectionModel cpadAccountsm1 = null;
   
   /** scroller for list of associated accounts */
   private JScrollPane cpadScroller2 = null;
   
   /** table to display associated accounts */
   private JTable cpadAccounts2 = null;
   
   /** related table model */
   private TaggedDataTableModel cpadAccounttm2 = null;
   
   /** related table selection model */
   private ListSelectionModel cpadAccountsm2 = null;

   /** All Subjects label */
   private JLabel cpadlA = null;
   
   /** All Subjects check-box */
   private JCheckBox cpadchkA = null;

   /** Add accounts button in modal dialog*/
   private JButton cpadbA = null;
   
   /** Remove accounts button in modal dialog*/
   private JButton cpadbR = null;
   
   /** Save button in modal dialog*/
   private JButton cpadbS = null;
   
   /** Close button in modal dialog*/
   private JButton cpadbC = null;
   
   /** modal dialog panel for change instance function */
   private JPanel cpadPanel = null;

   /** state definition - initial empty */
   private final boolean[] STATE_INITIAL_EMPTY =
   {
      /* center,                  */       true,
      /*    cpScroller,           */       true,
      /*       cpACLs,            */       true,
      /*    cpCtlPanel,           */       true,
      /*       cpbR,              */       true,
      /*       cpbA,              */       true,
      /*       cpbD,              */ false,
      /*       cpbC,              */ false,
      /*       cpbMu,             */ false,
      /*       cpbMd,             */ false,
      /*       cpbF               */       true,
      /*       cpbS               */ false
   };
   
   /** state definition - single selection */
   private final boolean[] STATE_ACL_SINGLE =
   {
      /* center,                  */       true,
      /*    cpScroller,           */       true,
      /*       cpACLs,            */       true,
      /*    cpCtlPanel,           */       true,
      /*       cpbR,              */       true,
      /*       cpbA,              */       true,
      /*       cpbD,              */       true,
      /*       cpbC,              */       true,
      /*       cpbMu,             */       true,
      /*       cpbMd,             */       true,
      /*       cpbF               */       true,
      /*       cpbS               */       true
   };

   /** state definition - multiple selection */
   private final boolean[] STATE_ACL_MULTI =
   {
      /* center,                  */       true,
      /*    cpScroller,           */       true,
      /*       cpACLs,            */       true,
      /*    cpCtlPanel,           */       true,
      /*       cpbR,              */       true,
      /*       cpbA,              */       true,
      /*       cpbD,              */       true,
      /*       cpbC,              */ false,
      /*       cpbMu,             */ false,
      /*       cpbMd,             */ false,
      /*       cpbF               */       true,
      /*       cpbS               */ false
   };
   
   /** All the ACLs, as received from the server. */
   private AclDef[] acls = null;

   /** All the ACLs, with no menu filters applied. */
   private List<AclRow> allAcls = null;
   
   /** Access Control filtered by account. */
   private String account;
   
   /** Access Control filtered by resource type. */
   private String resourceType;
   
   /** Access Control filtered by server. */
   private String serverName;
   
   /** The screen title. */
   private String screenTitle;
   
   /** The last applied filter */
   private String lastFilter;

   /** effective versus full view, when in specific resource type */
   private boolean effectiveView = false;
   
   /** subjects being added with the new ACL, array of names */
   private String[] subjects = null;
   
   /** subjects being added with the new ACL, array of indices */
   private int[] subjs = null;
   
   /** subjects being added with the new ACL, map */
   private Map<Integer, String> subjectsMap = null;

   /** instance of Rights being added with the new ACL */
   private Rights newRights = null;

   /** set to true while Add ACL dialog is active */
   private boolean addActive = false;

   /** set to true while changing selection of Effective view by code */
   private boolean changingEffective = false;

   /** Control standalone mode. */
   private final boolean standalone;

   /** Outer dialog for standalone mode. */
   private JDialog outerDialog = null;

   /**
    * Constructor.
    * 
    * @param    client
    *           instance of AdminClient
    */
   public AdminACL(AdminClient client)
   {
      this(client, false, null);
   }
   
   /**
    * Constructor.
    * 
    * @param    client
    *           instance of AdminClient
    * @param    standalone
    *           Prepare instance for work in standalone mode.
    * @param    filterText
    *           Initial filter text.
    */
   public AdminACL(AdminClient client, boolean standalone, String filterText)
   {
      // store useful references
      this.client = client;
      this.standalone = standalone;
      
      adm = client.getExports();
      
      // on build, set the 'live' server as server filter
      serverName = adm.getServerName();

      center = new JPanel();
      
      // create the main screen and add it to the swapper
      mainPanel = AdminUtils.groupComponents(new Component[][] {{center}});
      
      if (!standalone)
      {
         client.add(mainPanel, SCREEN_ACLS, this);
      }

      // top panel with group layout manages the table panel

      //  +--------center----------+
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    cpScroller      | |
      //  | |      CENTER        | |
      //  | |                    | |
      //  | |                    | |
      //  | +--------------------+ |
      //  | +--------------------+ |
      //  | |                    | |
      //  | |    cpCtlPanel      | |
      //  | |      SOUTH         | |
      //  | +--------------------+ |
      //  +------------------------+
      center.setLayout(new BorderLayout());

      cpACLtm = new AclTableModel();
      cpACLs = new JTable(cpACLtm);
      cpACLs.getTableHeader().setReorderingAllowed(false);
      cpACLs.setDefaultRenderer(Object.class, new AclTableRenderer());
      cpACLsm = cpACLs.getSelectionModel();
      cpACLsorter = new TableRowSorter<AclTableModel>(cpACLtm);

      if (standalone)
      {
         cpACLsorter.setRowFilter(RowFilter.regexFilter("^" + filterText + "$"));
         lastFilter = filterText;
      }
      else
      {
         cpACLsorter.setRowFilter(null);
      }
      
      cpACLs.setRowSorter(cpACLsorter);
      int num = cpACLs.getColumnCount();      
      for (int i = 0; i < num; i++)
      {
         TableColumn column = cpACLs.getColumnModel().getColumn(i);
         column.setPreferredWidth(cpACLtm.getColumnWidth(i));
      }
      
      cpScroller = new JScrollPane();
      cpScroller.getViewport().add(cpACLs);
      cpScroller.setBorder(new TitledBorder
         (new EtchedBorder(EtchedBorder.RAISED), " Defined Access Controls "));
      
      cpScroller.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      center.add(cpScroller, BorderLayout.CENTER);
      
      JPanel viewLine = new JPanel();
      cpcEv = new JCheckBox("Effective View");
      viewLine.setLayout(new BoxLayout(viewLine, BoxLayout.LINE_AXIS));
      viewLine.add(cpcEv);
      
      if (!standalone)
         center.add(viewLine, BorderLayout.NORTH);
                            
      // wpCtlPanel panel is controlled with GroupLayout
      //  +------wpCtlPanel--------+
      //  |                        |
      //  |                        |
      //  |     buttons here       |   
      //  |                        |
      //  |                        |
      //  +------------------------+
      
      // creating control widgets
      cpbCl = new JButton("Cancel");
      cpbR  = new JButton("Refresh");
      cpbA  = new JButton("Add ACL");
      cpbD  = new JButton("Delete Selected");
      cpbC  = new JButton("Clone ACL");
      cpbMu = new JButton("Move ACL Up");
      cpbMd = new JButton("Move ACL Down");
      cpbF  = new JButton("Filter");
      cpbS  = new JButton("Resequence ACLs");

      if (!standalone)
      {
         cpCtlPanel = AdminUtils.groupComponents(new Component[][] 
            { 
               { cpbR, cpbA, cpbD, cpbC, cpbMu, cpbMd, cpbF, cpbS }
            },
            true);
      }
      else
      {
         cpCtlPanel = AdminUtils.groupComponents(new Component[][] 
            { 
               { cpbCl, new JLabel(), cpbR, cpbA, cpbD, cpbC, cpbMu, cpbMd, cpbS }
            },
            true);
      }
      
      center.add(cpCtlPanel, BorderLayout.SOUTH);

      // create resequence ACLs dialog
      cpsdlId   = new JLabel("Starting Number");
      cpsdfId   = new JTextField(6);
      cpsdlStep = new JLabel("Increment By");
      cpsdfStep = new JTextField(6);
      cpsdbRe   = new JButton("Resequence");
      cpsdbC    = new JButton("Cancel");
      
      cpsPanel  = AdminUtils.groupComponents(new Component[][] {
               { cpsdlId,   cpsdfId },
               { cpsdlStep, cpsdfStep },
               { cpsdbC,    cpsdbRe },
         });

      // create the Change Instance Name dialog
      cprdlName    = new JLabel("Resource Instance Name");
      cprdfName    = new JTextField();
      cprdfName.setColumns(30);
      cprdlMode1   = new JLabel("Exact Nature");
      cprdchkMode1 = new JRadioButton();
      cprdlMode2   = new JLabel("Regexp Nature");
      cprdchkMode2 = new JRadioButton();
      
      ButtonGroup cprdbg = new ButtonGroup();
      cprdbg.add(cprdchkMode1);
      cprdbg.add(cprdchkMode2);
      
      cprdbS = new JButton("Save");
      cprdbC = new JButton("Cancel");
      
      cprPanel = AdminUtils.groupComponents(new Component[][] {
               { cprdlName,  cprdfName },
               { cprdlMode1, cprdchkMode1 },
               { cprdlMode2, cprdchkMode2 },
               { cprdbC,     cprdbS },
         }
      );
      
      // create add ACL item dialog
      cpddlInst    = new JLabel("Server this ACL item is for");
      cpddModI     = new DefaultComboBoxModel();
      cpddxInst    = new JComboBox(cpddModI);
      cpddxInst.setEditable(true);
      cpddlRtype   = new JLabel("Resource Type");
      cpddModR     = new DefaultComboBoxModel();
      cpddxRtype   = new JComboBox(cpddModR);
      cpddlId      = new JLabel("Access Control ID");
      cpddfId      = new JTextField(6);      
      cpddlName    = new JLabel("Resource Instance Name");
      cpddfName    = new JTextField(30);
      cpddlMode1   = new JLabel("Name Nature");
      cpddchkMode1 = new JRadioButton(" exact match");
      cpddlMode2   = new JLabel("");
      cpddchkMode2 = new JRadioButton(" regular expression");
      cpddlSubjs   = new JLabel("Subjects");
      cpddfSubjs   = new JTextField(30);
      cpddfSubjs.setFocusable(false);
      cpddlRights  = new JLabel("Rights");
      cpddfRights  = new JTextField(30);
      cpddfRights.setFocusable(false);
      
      ButtonGroup cpddbg = new ButtonGroup();
      cpddbg.add(cpddchkMode1);
      cpddbg.add(cpddchkMode2);
      
      cpddbS = new JButton("Save");
      cpddbC = new JButton("Cancel");
      
      cpddPanel = AdminUtils.groupComponents(new Component[][] 
         {
            { cpddlInst,  cpddxInst },
            { cpddlRtype, cpddxRtype },
            { cpddlId,    cpddfId },
            { cpddlName,  cpddfName },
            { cpddlMode1, cpddchkMode1 },
            { cpddlMode2, cpddchkMode2 },
            { cpddlSubjs, cpddfSubjs },
            { cpddlRights, cpddfRights },
            { cpddbC,     cpddbS },
         }
      );
      
      // create change instance name ACLs dialog
      cpidlI   = new JLabel("Private Instance");
      cpidfI   = new JTextField();
      cpidfI.setColumns(30);
      cpidlSh   = new JLabel("Shared Instance");
      cpidchkSh= new JCheckBox();
      
      cpidbS = new JButton("Save");
      cpidbC = new JButton("Cancel");
      
      cpiPanel = AdminUtils.groupComponents(new Component[][] {
               { cpidlI,  cpidfI },
               { cpidlSh, cpidchkSh },
               { cpidbC,  cpidbS },
         }
      );

      // create clone ACL dialog
      cpcdlId = new JLabel("Clone ID");
      cpcdfId = new JTextField(6);      
      cpcdbS  = new JButton("Save");
      cpcdbC  = new JButton("Cancel");
      
      cpcPanel = AdminUtils.groupComponents(new Component[][] {
            { cpcdlId, cpcdfId },
            { cpcdbC,  cpcdbS },
      }
      );
      
      // create ACL ID editing dialog
      cpedlId = new JLabel("Target ID");
      cpedfId = new JTextField(6);      
      cpedbS  = new JButton("Save");
      cpedbC  = new JButton("Cancel");
      
      cpePanel = AdminUtils.groupComponents(new Component[][] {
            { cpedlId, cpedfId },
            { cpedbC,  cpedbS },
      }
      );
      
      // create the manage subjects dialog
      {
         cpadAccounttm1 = new TaggedDataTableModel("Type", 
                                                   "Account", 
                                                   "Description");
         cpadAccounts1 = new JTable(cpadAccounttm1);
         cpadAccounts1.getTableHeader().setReorderingAllowed(false);
         cpadAccountsm1 = cpadAccounts1.getSelectionModel();        
         cpadScroller1 = new JScrollPane();
         cpadScroller1.getViewport().add(cpadAccounts1);
         cpadScroller1.setBorder(new TitledBorder
            (new EtchedBorder(EtchedBorder.RAISED), " Available Accounts "));

         cpadScroller1.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      }
      {
         cpadAccounttm2 = new TaggedDataTableModel("Type", 
                                                   "Account", 
                                                   "Description");
         cpadAccounts2 = new JTable(cpadAccounttm2);
         cpadAccounts2.getTableHeader().setReorderingAllowed(false);
         cpadAccountsm2 = cpadAccounts2.getSelectionModel();        
         cpadScroller2 = new JScrollPane();
         cpadScroller2.getViewport().add(cpadAccounts2);
         cpadScroller2.setBorder(new TitledBorder
            (new EtchedBorder(EtchedBorder.RAISED), " Associated Accounts "));
         
         cpadScroller2.setViewportBorder(new EmptyBorder(5, 5, 5, 5));
      }

      cpadlA = new JLabel("Associate All Accounts");
      cpadchkA = new JCheckBox();
      
      cpadbA = new JButton("Add Selected Accounts");
      cpadbR = new JButton("Remove Selected Accounts");

      cpadbS = new JButton("Save");
      cpadbC = new JButton("Cancel");

      {
         JPanel pnl0 = AdminUtils.groupComponents(new Component[][] 
                        {
                           { cpadlA, cpadchkA },
                        });
         JPanel pnl1 = AdminUtils.groupComponents(new Component[][] 
                        {
                           { cpadbA, cpadbR },
                           { cpadScroller1, cpadScroller2 }
                        });
         JPanel pnl2 = AdminUtils.groupComponents(
                             new Component[][] { { cpadbC, cpadbS } }, false);
         
         cpadPanel = AdminUtils.borderComponents(
                             new Component[] { pnl0, pnl1, pnl2 }, true);
      }
      
      // create event handlers: center panel

      // Effective View check box
      cpcEv.addItemListener(new ItemListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void itemStateChanged(ItemEvent evt) 
         {
            handleEffectiveView();
         }
      });

      // ACL table selection events
      cpACLsm.addListSelectionListener(new ListSelectionListener() 
      {
         public void valueChanged(ListSelectionEvent evt) 
         {
            handleACLsSelection(evt);
         }
      });

      // capture mouse clicks on table cell
      cpACLs.addMouseListener(new MouseAdapter()
      {
         public void mouseClicked(MouseEvent e)
         {
            if (e.getClickCount() == 2)
            {
               Point p = e.getPoint();
               int row = cpACLs.rowAtPoint(p);
               int column = cpACLs.columnAtPoint(p);

               handleACLCellEdit(row, column);
            }
         }
      });

      // Populate button
      cpbR.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handlePopulateACLs();
         }
      });

      // Add button
      cpbA.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleAddACL();
         }
      });

      // Delete Selected button
      cpbD.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleDeleteSelectedACL();
         }
      });

      // Clone ACL button
      cpbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCloneACLItem();
         }
      });

      // Move ACL Up button
      cpbMu.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleMoveACLUp();
         }
      });

      // Move ACL Down button
      cpbMd.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleMoveACLDown();
         }
      });

      // Filter ACLs button
      cpbF.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleFilter();
         }
      });

      // Resequence ACLs button
      cpbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleResequenceACLs();
         }
      });

      cpbCl.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            if (outerDialog != null)
            {
               outerDialog.setVisible(false);
               outerDialog.dispose();
            }
         }
      });

      
      // Resequence ACLs As button
      cpsdbRe.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleResequenceACLsAs();
         }
      });

      // Cancel button in Resequence ACLs As dialog
      cpsdbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // Change Resource Instance Name As button
      cprdbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleChangeResourceName();
         }
      });

      // Cancel button in change Resource Instance Name As dialog
      cprdbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // Selection of resource type in the Add ACL Item dialog
      cpddxRtype.addActionListener(new ActionListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleResourceType();
         }
      });

      // capture mouse clicks on subjects field in the Add ACL Item dialog
      cpddfSubjs.addMouseListener(new MouseAdapter()
      {
         public void mouseClicked(MouseEvent e)
         {
            if (e.getClickCount() == 2)
            {
               handleAddSubjects();
            }
         }
      });

      // capture mouse clicks on rights field in the Add ACL Item dialog
      cpddfRights.addMouseListener(new MouseAdapter()
      {
         public void mouseClicked(MouseEvent e)
         {
            if (e.getClickCount() == 2)
            {
               handleAddRights();
            }
         }
      });

      // Save button in Add ACL Item dialog
      cpddbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleSaveACL();
         }
      });

      // Cancel button in Add ACL Item dialog
      cpddbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // Clone ACL item button
      cpcdbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCloneACLItemAs();
         }
      });

      // Cancel button in Clone ACLdialog
      cpcdbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // direct ID editing dialog - Save button
      cpedbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleMoveAclAround();
         }
      });

      // direct ID editing dialog - Cancel button
      cpedbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // Shared check-box in Change Instance As button
      cpidchkSh.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            cpidfI.setEnabled(!cpidchkSh.isSelected());
         }
      });

      // Change Instance As button
      cpidbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleChangeInstance();
         }
      });

      // Cancel button in change Instance As dialog
      cpidbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // All Accounts check box
      cpadchkA.addItemListener(new ItemListener()
      {
         @Override
         @SuppressWarnings("unused") 
         public void itemStateChanged(ItemEvent evt) 
         {
            handleAllAccounts();
         }
      });

      // Double-click in Available Accounts table
      cpadAccounts1.addMouseListener(new MouseAdapter()
      {
         public void mouseClicked(MouseEvent e)
         {
            if (e.getClickCount() == 2)
            {
               handleSubjectSelection(true);
            }
         }
      });

      // Double-click in Associated Accounts table
      cpadAccounts2.addMouseListener(new MouseAdapter()
      {
         public void mouseClicked(MouseEvent e)
         {
            if (e.getClickCount() == 2)
            {
               handleSubjectSelection(false);
            }
         }
      });

      // Add Subjects button
      cpadbA.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleSubjectSelection(true);
         }
      });

      // Remove Subjects button
      cpadbR.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleSubjectSelection(false);
         }
      });

      // Change Subjects to button
      cpadbS.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleChangeSubjects();
         }
      });

      // Cancel button in change Subjects To dialog
      cpadbC.addActionListener(new ActionListener() 
      {
         @Override
         @SuppressWarnings("unused") 
         public void actionPerformed(ActionEvent evt) 
         {
            handleCancel();
         }
      });

      // create the array of components for the state machine
      state = new JComponent[]
      {
         center,
            cpScroller,
               cpACLs,
            cpCtlPanel,
               cpbR, cpbA,  cpbD, cpbC, cpbMu, cpbMd, cpbF, cpbS
      };

      // set initial state of all components
      setACLsNoSelection();
   }

   /**
    * Get main panel.
    * 
    * @return   main panel in standalone mode and <code>null</code> otherwise.
    */
   public JPanel getMainPanel()
   {
      if (standalone)
         return mainPanel;
      
      return null;
   }
   
   /**
    * Set reference to outer dialog.
    * 
    * @param    dialog
    */
   public void setDialog(JDialog dialog)
   {
      outerDialog = dialog;
   }
   
   /**
    * Performs printing of the current screen data. The ACL screen prints the
    * current selection of the ACL table.
    */
   public void print()
   {
      MessageFormat header = new MessageFormat(screenTitle);
      MessageFormat footer = new MessageFormat("page {0, number}");
      
      try
      {
         cpACLs.print(JTable.PrintMode.FIT_WIDTH, header, footer);
      }
      catch (PrinterException pe)
      {
         client.beep("printer exception" + pe);
      }
      catch (SecurityException se)
      {
         // nothing to do
      }
   }
   
   /**
    * Queries the extended printing capability. If the extended capability is
    * not provided, the printExtended() method should not be called.
    * 
    * @return   <code>false</code> to tell the caller the extended printing is
    *           not implemented.
    */
   public boolean canPrintExtended()
   {
      return false;
   }

   /**
    * Does nothing here.
    */
   public void printExtended()
   {
   }
   
   /**
    * Show the ACLs by account.
    */
   void showByAccount()
   {
      // get and validate the account name
      final int type = JOptionPane.QUESTION_MESSAGE;
      final String txt = "Enter an account name:";
      final String title = "Account name";
      String acc = JOptionPane.showInputDialog(client,
                                               txt,
                                               title,
                                               type);
      if (acc == null)
      {
         return;
      }
      
      acc = acc.trim();
      
      AdminUtils.busy(client);
      boolean found = "all_others".equals(acc)    ||
                      adm.getUser(acc)    != null || 
                      adm.getProcess(acc) != null || 
                      adm.getGroup(acc)   != null;
      AdminUtils.idle(client);

      if (!found)
      {
         AdminUtils.errorDialog(client, 
                                "Error !", 
                                "The account <" + acc + "> is invalid !");
         showByAccount();
      }

      showByAccount(acc);
   }
   
   /**
    * Show the ACLs for the specified account, using the live server.
    * 
    * @param    acc
    *           The account used to filter the the ACLs.
    */
   void showByAccount(String acc)
   {
      // busy
      AdminUtils.busy(client);

      account = acc;
      switchResourceType(null);

      screenTitle = TITLE_ACLS + " for account <" + account + ">" +
                    " and server <" + serverName + ">";
      showACLs();

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Show all ACLs.
    */
   void showAll()
   {
      // busy
      AdminUtils.busy(client);

      // remove all filters and sorting
      account = null;
      switchResourceType(null);
      
      screenTitle = TITLE_ACLS + " for server <" + serverName + ">";
      showACLs();

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Show the ACLs by resource type.
    * 
    * @param    resourceType
    */
   public void showByResourceType(String resourceType)
   {
      // busy
      AdminUtils.busy(client);

      // set the resource type as filter and remove any other filters
      serverName = adm.getServerName();
      account = null;
      switchResourceType(resourceType);

      screenTitle = TITLE_ACLS + " for resource <" + resourceType + ">" +
                    " and server <" + serverName + ">";
      showACLs();

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Show the ACLs by server.
    * 
    * @param    server
    */
   void showByServer(String server)
   {
      // busy
      AdminUtils.busy(client);

      this.serverName = server;
      this.account = null;
      switchResourceType(null);

      screenTitle = TITLE_ACLS + " for server <" + server + ">";
      showACLs();

      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Show the ACL screen with the computed {@link #screenTitle}.
    */
   private void showACLs()
   {
      if (!standalone)
      {
         client.show(SCREEN_ACLS, screenTitle);

         // remove any manual filter
         lastFilter = null;
         cpACLsorter.setRowFilter(null);
      }
      doPopulateACLs();
   }
   
   /**
    * Handle "Effective View" check box state change. The reaction is to
    * repopulate the table and refresh the view.
    */
   private void handleEffectiveView()
   {
      // protection against unwanted handler invocation
      if (changingEffective)
      {
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // populate the table afresh
      populateACLList();
      
      // idle
      AdminUtils.idle(client);
}
   
   /**
    * Handle row selection change in the ACL table.
    * 
    * @param    evt
    *           event caused this method call
    */
   private void handleACLsSelection(ListSelectionEvent evt)
   {
      // do not bother until the dust is fully settled
      if (evt.getValueIsAdjusting())
      {
         return;
      }
      
      // figure out the selection state
      if (cpACLsm.isSelectionEmpty())
      {
         // no selection
         setACLsNoSelection();
      }
      else if (cpACLs.getSelectedRows().length == 1)
      {
         // single selection
         setACLsSingleSelection();
      }
      else
      {
         // multiple selection
         setACLsMultiSelection();
      }
   }
   
   /**
    * Move the subjects from a table to another, depending on action (add or
    * remove).
    * 
    * @param    add
    *           <code>true</code> if the selected accounts in the available
    *           accounts table are moved to the associated accounts table.
    */
   private void handleSubjectSelection(boolean add)
   {
      TaggedDataTableModel tmSource = add ? cpadAccounttm1 : cpadAccounttm2;
      ListSelectionModel smSource = add ? cpadAccountsm1 : cpadAccountsm2;
      JTable tSource = add ? cpadAccounts1 : cpadAccounts2;

      TaggedDataTableModel tmDest = add ? cpadAccounttm2 : cpadAccounttm1;
      
      int[] sele = tSource.getSelectedRows();
      if (sele.length == 0)
      {
         client.beep("select an account first");
         return;
      }
      
      smSource.clearSelection();
      
      // sort the array of selected indices to walk backwards
      Arrays.sort(sele);
      
      for (int i = sele.length - 1; i >= 0; i--)
      {
         int sel = tSource.convertRowIndexToModel(sele[i]);
         AccountRow row = (AccountRow) tmSource.data.get(sel);
         tmSource.deleteRow(sel);
         tmDest.addRow(row);
      }
   }
   
   /**
    * Handle Add ACL Item button press. Prepares and displays "Add ACL" dialog.
    */
   private void handleAddACL()
   {
      // enumerate all ACL instances and stuff the combo box
      String[] inst = adm.listAclInstances();
      cpddModI.removeAllElements();
      
      for (int i = 0; i < inst.length; i++)
      {
         String in = inst[i].length() == 0 ? SHARED_INSTANCE : inst[i];
         cpddModI.addElement(in);
      }
      
      // enumerate all resource types and stuff the combo box
      String[] type = client.getResourceTypes();
      cpddModR.removeAllElements();
      
      for (int i = 0; i < type.length; i++)
      {
         cpddModR.addElement(type[i]);
      }
      
      // if there is a selection, assign better defaults based on the first
      // selected row
      int sel = cpACLs.getSelectedRow();
      int id = -1;
      
      if (sel >= 0)
      {
         // compute a better new id
         id = nextId(false, true);
      }
      else
      {
         // use the first row in the viewport as a reference
         sel = 0;
      }
      
      if (cpACLtm.data.size() > 0 && cpACLs.getRowCount() > 0)
      {
         sel = cpACLs.convertRowIndexToModel(sel);
         AclRow row = (AclRow)cpACLtm.data.get(sel);
         
         // set better default ACL instance and resource type
         String selInstance = row.getInstance();
         String in = selInstance.length() == 0 ? SHARED_INSTANCE : selInstance;
         cpddModI.setSelectedItem(in);
         
         String selResourceType = row.getResourceType();
         cpddModR.setSelectedItem(selResourceType);
      }
      else
      {
         cpddxInst.setSelectedIndex(0);
         cpddxRtype.setSelectedIndex(0);
      }
      
      // reset fields
      cpddfId.setText(id == -1 ? "" : Integer.toString(id));
      cpddfName.setText("");
      cpddchkMode1.setSelected(true);
      cpddchkMode2.setSelected(false);
      cpddfSubjs.setText("all_others");

      if (standalone)
      {
         cpddModR.setSelectedItem(resourceType);
         cpddfName.setText(lastFilter);
         cpddfName.setEnabled(false);
         cpddxRtype.setEnabled(false);
      }
      
      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "Add ACL item", 
                                      cpddPanel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }      
   }
   
   /**
    * Handle selection of resource type in the Add ACL item dialog.
    */
   private void handleResourceType()
   {
      // get the current selection in the resource type combo
      String resourceType = (String)cpddxRtype.getSelectedItem();
      
      // create a default instance of rights
      RightsEditor editor = client.getCustomAclEditor(resourceType);
      
      if (editor != null)
      {
         newRights = editor.createDefaultRights();
         cpddfRights.setText(newRights.toString());
      }
      else
      {
         newRights = null;
         cpddfRights.setText("");
      }
   }
   
   /**
    * Handle double click on subjects in the Add ACL item dialog
    * by invoking Subjects Editor.
    */
   private void handleAddSubjects()
   {
      if (subjects == null)
      {
         subjects = new String[] {"all_others"};
         subjs = new int[] { AclDef.ALL_OTHERS };
      }
      
      String[] newSubjs = editSubjects(subjects);
      if (newSubjs != null)
      {
         subjects = newSubjs;
         StringBuilder sb = new StringBuilder();
         for (int i = 0; i < subjects.length; i++)
         {
            sb.append(subjects[i]);
            if (i < subjects.length - 1)
            {
               sb.append(",");
            }
         }
         cpddfSubjs.setText(sb.toString());
      }
   }
   
   /**
    * Handle double click on rights in the Add ACL item dialog
    * by invoking Rights Editor.
    */
   private void handleAddRights()
   {
      // get the current selection in the resource type combo
      String resourceType = (String)cpddxRtype.getSelectedItem();
      RightsEditor editor = client.getCustomAclEditor(resourceType);

      // get the resource instance name and nature
      String resourceName = cpddfName.getText();
      boolean exact = cpddchkMode1.isSelected();
      
      // edit rights
      Rights rights = editor.edit(resourceName, exact, newRights);

      if (rights != null)
      {
         newRights = rights;
         cpddfRights.setText(newRights.toString());
      }
   }
   
   /**
    * Handle Save button press in the Add ACL item dialog.
    */
   private void handleSaveACL()
   {
      // get the ACL instance name from the dialog
      String instance = (String)cpddxInst.getSelectedItem();
      if (instance.length() == 0)
      {
         // set initial focus on the field
         cpddxInst.requestFocusInWindow();
         
         // alarm
         client.beep("select or type in the ACL instance");
         return;
      }
      
      if (instance.equalsIgnoreCase(SHARED_INSTANCE))
      {
         instance = "";
      }
      
      // validate the new ID
      Integer id = validateId(cpddfId, "not a valid numeric ID");
      if (id == null)
      {
         return;
      }
      
      // warn about creation of a new ACL instance
      if (cpddxInst.getSelectedIndex() == -1)
      {
         if (!AdminUtils.confirmDialog(client, "Warning",
             String.format("About to create a new ACL instance <%s>. Continue?",
                           instance)))
         {
            handleCancel();
            return;
         }

      }
      
      // get the resource type from the dialog
      String resourceType = (String)cpddxRtype.getSelectedItem();
      
      // validate the resource instance name
      String nm = validateStringField(cpddfName, 
                                      "enter a Resource Instance Name");
      if (nm == null)
      {
         return;
      }

      // build a new ACL
      AclDef ad    = new AclDef();
      AclDef.Acl acl = ad.new Acl();
      acl.origin   = instance;
      acl.oid      = id;
      acl.name     = nm;
      acl.exact    = cpddchkMode1.isSelected();
      acl.subjects = subjs;
      acl.rights = newRights;
      
      // busy
      AdminUtils.busy(client);
      
      // attempt addition
      boolean rc = adm.addAcl(resourceType, 
                              acl, 
                              subjectsMap);
      if (!rc)
      {
         // attempt has failed
         client.alarm();
         
         // idle
         AdminUtils.idle(client);
         
         return;
      }
      
      // attempt has succeeded; refresh row
      client.ok();
      
      populateACLList();
      
      // locate the new ACL and make it visible and selected
      int sel = locateAcl(resourceType, instance, id);
      if (sel >= 0)
      {
         sel = cpACLs.convertRowIndexToView(sel);
         cpACLsm.setSelectionInterval(sel, sel);
         ensureVisibility(sel);
      }
      
      // reset the subjects related data
      subjects = null;
      subjs = null;
      subjectsMap = null;

      handleCancel();

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Handle Delete Selected ACL button press.
    */
   private void handleDeleteSelectedACL()
   {
      int[] sele = cpACLs.getSelectedRows();
      
      // safety check
      if (sele.length == 0)
      {
         client.beep("select some items first");
         return;
      }
      
      // ask confirmation
      String quest = null;
      String title = null;
      
      if (sele.length == 1)
      {
         int sel = cpACLs.convertRowIndexToModel(sele[0]);
         AclRow row = (AclRow) cpACLtm.data.get(sel);
         
         quest = String.format("Deleting ACL %d in <%s><%s>. Continue?",
                               row.getId(), row.getValueAt(0), 
                               row.getValueAt(1));
         title = "Deleting item";
      }
      else
      {
         quest = String.format("Deleting %d items. Continue?", sele.length);
         title = "Deleting items";
      }
      
      if (!AdminUtils.confirmDialog(client, title, quest))
      {
         // declined
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // do not keep the selection
      cpACLs.clearSelection();
      
      // sort the array of selected indices to walk backwards
      Arrays.sort(sele);
      
      boolean rrc = true;
      
      for (int i = sele.length - 1; i >= 0; i--)
      {
         int sel = cpACLs.convertRowIndexToModel(sele[i]);
         AclRow row = (AclRow) cpACLtm.data.get(sel);
   
         // attempt deletion
         boolean rc = adm.deleteAcl(row.getResourceType(), 
                                    row.getInstance(), 
                                    row.getId());
         if (rc)
         {
            // attempt has succeeded; delete this row
            cpACLtm.deleteRow(sel);
         }
         else
         {
            rrc = false;
         }
      }
      
      if (!rrc)
      {
         // attempt has failed
         client.alarm();
      }
      else
      {
         // attempt has succeeded; delete this row
         client.ok();
      }

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Handle mouse double-click in a table cell. This allows edit for the
    * following cells:
    * <ul>
    *    <li>instance name - to move the ACL item to another instance.</li>
    *    <li>resource name and nature (exact or regexp)</li>
    *    <li>subjects list</li>
    *    <li>rights</li>
    * </ul>
    */
   private void handleACLCellEdit(int sel, int cell)
   {
      // safety check
      if (sel == -1)
      {
         return;
      }
      
      sel = cpACLs.convertRowIndexToModel(sel);

      AclRow aclItem = (AclRow) cpACLtm.data.get(sel);

      String title = "";
      JPanel panel = null;
      
      if (cell == 0)
      {
         // edit instance
         title = "Move ACL to another Instance";
         panel = cpiPanel;

         // load the data
         cpidchkSh.setSelected("".equals(aclItem.getInstance()));

         cpidfI.setText(aclItem.getInstance());
         cpidfI.setEnabled(!cpidchkSh.isSelected());
      }

      if (cell == 2)
      {
         // edit ID
         title = "Move ACL around";
         panel = cpePanel;
         cpedfId.setText("");
      }

      if (cell == 3)
      {
         // edit resource instance name and mode (regexp or not)
         title = "Change Resource Instance Name";
         panel = cprPanel;
         
         // load the data
         if (!aclItem.isErrorName())
         {
            cprdfName.setText(aclItem.getName());
         }
         else
         {
            cprdfName.setText("");
         }
         
         cprdchkMode1.setSelected(aclItem.getExact());
         cprdchkMode2.setSelected(!aclItem.getExact());
      }

      if (cell == 4)
      {
         // edit subjects
         title = "Change Associated Subjects";
         panel = cpadPanel;
         
         String[] subjects = aclItem.subjectsSet.toArray(new String[]{});
         if (aclItem.acl.subjects.length == 1 && 
             aclItem.acl.subjects[0] == AclDef.ALL_OTHERS)
         {
            subjects = new String[] {"all_others"};
         }
         
         if (!loadSubjects(subjects))
         {
            return;
         }
      }

      if (cell == 5)
      {
         handleChangeRights();
         return;
      }
      
      if (panel == null)
      {
         return;
      }
      
      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      title, 
                                      panel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }
   
   /**
    * Edits the list of subjects as part of Add ACL dialog.
    * 
    * @param    subjs
    *           subject IDs to edit
    * 
    * @return   array of subject IDs or <code>null</code>
    */
   private String[] editSubjects(String[] subjs)
   {
      // load dialog panels with account data
      if (!loadSubjects(subjs))
      {
         return null;
      }
      
      subjects = null;
      
      // display nested modal dialog
      JDialog saveDialog = dialog;
      addActive = true;
      
      dialog = AdminUtils.modalDialog(client, 
                                      "Change Associated Subjects",
                                      cpadPanel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
      dialog = saveDialog;
      addActive = false;
      
      return subjects;
   }
   
   /**
    * Load the subjects in the dialog.
    * 
    * @return   <code>true</code> if loading was successfully.
    */
   private boolean loadSubjects(String[] subjs)
   {
      Set<String> subjectsSet = new TreeSet<String>();
      boolean allOthers = false;
      
      if (subjs.length == 1 && subjs[0].equalsIgnoreCase("all_others"))
      {
         allOthers = true;
      }
      else
      {   
         for (int i = 0; i < subjs.length; i++)
         {
            subjectsSet.add(subjs[i]);
         }
      }

      AdminUtils.busy(client);

      TaggedName[] users = adm.listUsers();
      if (users == null)
      {
         AdminUtils.idle(client);
         client.alarm();
         return false;
      }
      
      TaggedName[] groups = adm.listGroups();
      if (groups == null)
      {
         AdminUtils.idle(client);
         client.alarm();
         return false;
      }

      TaggedName[] procs = adm.listProcesses();
      if (procs == null)
      {
         AdminUtils.idle(client);
         client.alarm();
         return false;
      }
      
      List<AccountRow> accounts = new ArrayList<AccountRow>();
      accounts.addAll(createAccountRows("group", groups));
      accounts.addAll(createAccountRows("process", procs));
      accounts.addAll(createAccountRows("user", users));
      
      // load the accounts and select the associated ones
      cpadAccountsm1.clearSelection();
      cpadAccountsm2.clearSelection();
      cpadAccounttm1.clear();
      cpadAccounttm2.clear();

      // remove associated subjects
      Iterator<AccountRow> itr = accounts.iterator();
      while (itr.hasNext())
      {
         AccountRow acc = itr.next();
         if (subjectsSet.contains(acc.getName()))
         {
            itr.remove();
         }
      }
      
      // create the associated subjects rows
      List<TaggedName> userSet = new ArrayList<TaggedName>(
                                                     Arrays.asList(users));
      List<TaggedName> procSet = new ArrayList<TaggedName>(
                                                     Arrays.asList(procs));
      List<TaggedName> groupSet = new ArrayList<TaggedName>(
                                                    Arrays.asList(groups));
      
      TaggedName taggedName = new TaggedName("", "");
      List<AccountRow> assocAccounts = new ArrayList<AccountRow>();
      for (String acc : subjectsSet)
      {
         String type = "";
         TaggedName row = taggedName;
         row.setName(acc);

         if (userSet.contains(row))
         {
            type = "user";
            row = userSet.get(userSet.indexOf(row));
         }
         else if (groupSet.contains(row))
         {
            type = "group";
            row = groupSet.get(groupSet.indexOf(row));
         }
         else if (procSet.contains(row))
         {
            type = "process";
            row = procSet.get(procSet.indexOf(row));
         }
         else
         {
            type = "*not found*";
            row.setTag("");
         }
         
         assocAccounts.add(new AccountRow(type, row));
      }
      
      cpadAccounttm1.loadAll(accounts.toArray(new AccountRow[0]));      
      cpadAccounttm2.loadAll(assocAccounts.toArray(new AccountRow[0]));      

      // conditionally enable/disable subjects panels
      cpadchkA.setSelected(allOthers);
      handleAllAccounts();
      
      AdminUtils.idle(client);
      
      return true;
   }

   /**
    * Create the rows corresponding with the given account list.
    * 
    * @param    type
    *           the type of this accounts
    * @param    accounts
    *           the account list
    * 
    * @return   accounts added to a list of {@link AccountRow} objects.
    */
   private List<AccountRow> createAccountRows(String type, 
                                              TaggedName[] accounts)
   {
      List<AccountRow> rows = new ArrayList<AccountRow>();
      for (int i = 0; i < accounts.length; i++)
      {
         rows.add(new AccountRow(type, accounts[i]));
      }
      
      return rows;
   }
   
   /**
    * Handle rights change for the selected ACL item.
    */
   @SuppressWarnings("unchecked")
   private void handleChangeRights()
   {
      int sel = cpACLs.getSelectedRow();

      // safety check
      if (sel == -1)
      {
         return;
      }

      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);

      // get the resource instance name and nature
      String resourceName = row.getName();
      boolean exact = row.getExact();
      
      // edit rights
      RightsEditor editor = client.getCustomAclEditor(row.getResourceType());
      Rights source = row.acl.rights;
      if (row.isErrorRights())
      {
         source = editor.createDefaultRights();
      }
      Rights rights = editor.edit(resourceName, exact, source);

      if (rights != null)
      {
         row.acl.rights = rights;
         row.clearErrorRights();

         // the new rights need to be saved
         // busy
         AdminUtils.busy(client);
         
         // ensure instance name is provided
         boolean setOrigin = false;
         if (row.acl.origin == null)
         {
            setOrigin = true;
            row.acl.origin = row.aclDef.server;
         }
         
         // attempt addition
         boolean rc = adm.setAcl(row.getResourceType(), 
                                 row.acl, 
                                 Collections.EMPTY_MAP, 
                                 false, 
                                 false, 
                                 false, 
                                 true);
         if (setOrigin)
         {
            row.acl.origin = null;
         }
         
         if (!rc)
         {
            // attempt has failed
            client.alarm();
         }
         else
         {
            // attempt has succeeded; refresh row
            client.ok();
            cpACLtm.refreshRow(row, sel);
         }
   
         // idle
         AdminUtils.idle(client);
      }
   }
   
   /**
    * Change the resource instance name and exact nature.
    */
   @SuppressWarnings("unchecked")
   private void handleChangeResourceName()
   {
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         return;
      }
      String nm = validateStringField(cprdfName, 
                                      "Enter a Resource Instance Name !");
      if (nm == null)
      {
         return;
      }

      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);

      // busy
      AdminUtils.busy(client);
      
      AclDef.Acl acl = row.acl;
      acl.name   = nm;
      acl.exact  = cprdchkMode1.isSelected();
      
      // ensure instance name is provided
      boolean setOrigin = false;
      if (acl.origin == null)
      {
         setOrigin = true;
         acl.origin = row.aclDef.server;
      }
      
      // attempt change
      boolean rc = adm.setAcl(row.getResourceType(), 
                              acl, 
                              Collections.EMPTY_MAP,
                              true, 
                              true, 
                              false, 
                              false);
      if (setOrigin)
      {
         acl.origin = null;
      }
      
      // idle
      AdminUtils.idle(client);

      if (!rc)
      {
         // attempt has failed
         client.alarm();
         return;
      }
      
      // attempt has succeeded; refresh row
      client.ok();
      row.setName(nm);
      row.setExact(acl.exact);
      row.clearErrorName();
      row.clearErrorExact();
      cpACLtm.refreshRow(row, sel);

      handleCancel();
   }
   
   /**
    * Move the ACL item to another instance (shared or private).
    */
   private void handleChangeInstance()
   {
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         return;
      }
      
      boolean shared = cpidchkSh.isSelected();
      String newInstance = "";
      
      if (!shared)
      {
         newInstance = validateStringField(cpidfI, 
                                  "Enter a Private Instance !");
         if (newInstance == null)
         {
            return;
         }
      }

      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);

      // busy
      AdminUtils.busy(client);
      // attempt change
      boolean rc = adm.moveAcl(row.getResourceType(), 
                               row.acl.origin, 
                               row.acl.oid, 
                               newInstance);
      if (!rc)
      {
         // idle
         AdminUtils.idle(client);
         
         // attempt has failed
         client.alarm();
         return;
      }
      
      // attempt has succeeded; refresh all rows
      client.ok();
      doPopulateACLs();

      handleCancel();
      
      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Handle "Associate All Accounts" check box state change.
    * The reaction is to enable or disable account tables and buttons.
    */
   private void handleAllAccounts()
   {
      boolean state = !cpadchkA.isSelected();

      cpadAccounts1.setEnabled(state);
      cpadAccounts2.setEnabled(state);
      cpadbA.setEnabled(state);
      cpadbR.setEnabled(state);
   }
   
   /**
    * Set the associated accounts for the selected ACL item.
    * 
    * @return   concatenated subjects
    */
   @SuppressWarnings("unchecked")
   private String saveSubjects()
   {
      // get all selected accounts
      List<String> accounts = new ArrayList<String>();
      StringBuilder subjects = new StringBuilder();
      subjectsMap = Collections.EMPTY_MAP;
      
      if (!cpadchkA.isSelected())
      {
         for (TaggedName tn : cpadAccounttm2.data)
         {
            accounts.add(tn.getName());
         }
         
         if (accounts.size() == 0)
         {
            client.beep("select at least one account");
            return "";
         }
         
         // build a new array of accounts
         this.subjs = new int[accounts.size()];
         this.subjects = new String[accounts.size()];
         subjectsMap = new HashMap<Integer, String>();

         // build the account dictionary
         for (int i = 0; i < accounts.size(); i++)
         {
            subjectsMap.put(i, accounts.get(i));
            subjs[i] = i;
            this.subjects[i] = accounts.get(i);
            subjects.append(accounts.get(i));
            
            if (i < accounts.size() - 1)
            {
               subjects.append(", ");
            }
         }
      }
      else
      {
         this.subjs = new int[] {AclDef.ALL_OTHERS};
         this.subjects = new String[] {"all_others"};
         subjects.append("all_others");
      }
      
      return subjects.toString();
   }
   
   /**
    * Set the associated accounts for the selected ACL item.
    */
   private void handleChangeSubjects()
   {
      // get the results from the dialog
      String allsubjs = saveSubjects();
      
      // check for no selection
      if (allsubjs.length() == 0)
      {
         return;
      }
      
      // add ACL versus edit 
      if (addActive)
      {
         handleCancel();
         return;
      }
      
      // put changes into the ACL
      int sel = cpACLs.getSelectedRow();

      // safety check
      if (sel == -1)
      {
         return;
      }
      
      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);
      row.acl.subjects = subjs;
      
      // busy
      AdminUtils.busy(client);
      
      // ensure instance name is provided
      boolean setOrigin = false;
      if (row.acl.origin == null)
      {
         setOrigin = true;
         row.acl.origin = row.aclDef.server;
      }
      
      // attempt change
      boolean rc = adm.setAcl(row.getResourceType(), 
                              row.acl,
                              subjectsMap, 
                              false, 
                              false, 
                              true, 
                              false);
      if (setOrigin)
      {
         row.acl.origin = null;
      }
      
      // idle
      AdminUtils.idle(client);

      if (!rc)
      {
         // attempt has failed
         client.alarm();
         return;
      }
      
      // attempt has succeeded; refresh row
      client.ok();
      row.subjects = allsubjs;
      row.subjectsSet = new TreeSet<String>(subjectsMap.values());
      row.clearErrorSubjs();
      cpACLtm.refreshRow(row, sel);
      
      handleCancel();
   }
   
   /**
    * Handle Clone ACL button press.
    */
   private void handleCloneACLItem()
   {
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         client.beep("select an ACL item first");
         return;
      }

      sel = cpACLs.convertRowIndexToModel(sel);

      // compute the new ID
      int id = nextId(false, true);
      cpcdfId.setText(id == -1 ? "" : Integer.toString(id));

      // focus on start id
      cpsdfId.requestFocusInWindow();

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "Clone ACL Item", 
                                      cpcPanel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }

   /**
    * Move the ACL item up or down.
    * 
    * @param    up
    *           <code>true</code> if the item needs to be moved up.
    */
   private void moveACLItem(boolean up)
   {
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         return;
      }

      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);
      
      // compute a new ID candidate
      int newId = nextId(up, false);
      if (newId == -1)
      {
         client.beep("no room to move to");
         return;
      }

      // move without confirmation
      moveACL(row, newId);
      
      // locate the relocated ACL and make it visible and selected
      sel = locateAcl(row.getResourceType(), row.getInstance(), newId);
      if (sel >= 0)
      {
         sel = cpACLs.convertRowIndexToView(sel);
         cpACLsm.setSelectionInterval(sel, sel);
         ensureVisibility(sel);
      }
   }

   /**
    * Move an ACL to a directly specified position.
    * 
    * @param    row
    *           the ACL to move
    * 
    * @param    newId
    *           new position for the ACL
    */
   private void moveACL(AclRow row, int newId)
   {
      // busy
      AdminUtils.busy(client);
      
      // attempt movement
      boolean rc = adm.moveAcl(row.getResourceType(), 
                               row.getInstance(), 
                               row.getId(),
                               newId);
      if (!rc)
      {
         // attempt has failed
         client.alarm();
      }
      else
      {
         // attempt has succeeded; refresh table
         client.ok();
         
         populateACLList();
         
         // locate the relocated ACL and make it visible and selected
         int sel = locateAcl(row.getResourceType(), row.getInstance(), newId);
         if (sel >= 0)
         {
            sel = cpACLs.convertRowIndexToView(sel);
            cpACLsm.setSelectionInterval(sel, sel);
            ensureVisibility(sel);
         }
      }

      // idle
      AdminUtils.idle(client);
   }

   /**
    * Compute another ID based on the ID of the selected item.
    * 
    * @param    up
    *           <code>true</code> if the new ID must be lower than the item's
    *           ID. <code>false</code> if the ID must be greater than the
    *           item's ID.
    * @param    current
    *           <code>true</code> if the selected item is considered the
    *           origin. <code>false</code> if the prior/next item is used,
    *           depending on the direction (up/down).
    * 
    * @return   the new ID
    */
   private int nextId(boolean up, boolean current)
   {
      // figure out initially selected row
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         return -1;
      }
           
      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow selRow = (AclRow)cpACLtm.data.get(sel);
      
      // check to see whether the selected row is the boundary
      int dir = (up ? -1 : 1);
      
      int selIdx = allAcls.indexOf(selRow);
      int sid = selRow.getId();
      
      boolean boundary = false;
      AclRow immedNext = null;
      int immedIdx = -1;
      int iid = -1;
      
      if (selIdx == 0 && up)
      {
         boundary = true;
      }
      else if (selIdx == allAcls.size() - 1 && !up)
      {
         boundary = true;
      }
      else
      {
         immedIdx = selIdx + dir;
         immedNext = (AclRow)cpACLtm.data.get(immedIdx);
         iid = immedNext.getId();
         
         if (!immedNext.getInstance().equals(selRow.getInstance()) ||
             !immedNext.getResourceType().equals(selRow.getResourceType()))
         {
            boundary = true;
         }
      }
      
      // handle the boundary case
      if (boundary)
      {
         if (up)
         {
            if (sid == 1)
            {
               // no room to move to
               return -1;
            }

            return sid / 2;
         }

         return sid + 100; 
      }
      
      // when moving, the next item is used (current flag is false)
      // when adding/cloning, the current item must be used
      
      if (current)
      {
         //  place the new ID between the initial selection and immediate next
         int distance = Math.abs(iid - sid);
         if (distance == 1)
         {
            // no room to move to
            return -1;
         }

         return sid + dir * distance / 2;
      }
      
      // check to see whether the immediate next row is the boundary
      int nextIdx = -1;
      AclRow nextRow = null;
      int nid = -1;
      
      if (immedIdx == 0 && up)
      {
         boundary = true;
      }
      else if (immedIdx == allAcls.size() - 1 && !up)
      {
         boundary = true;
      }
      else
      {
         nextIdx = immedIdx + dir;
         nextRow = (AclRow)cpACLtm.data.get(nextIdx);
         nid = nextRow.getId();
         
         if (!nextRow.getInstance().equals(selRow.getInstance()) ||
             !nextRow.getResourceType().equals(selRow.getResourceType()))
         {
            boundary = true;
         }
      }
      
      // handle the immediate next boundary case
      if (boundary)
      {
         if (up)
         {
            if (iid == 1)
            {
               // no room to move to
               return -1;
            }

            return iid / 2;
         }

         return iid + 100; 
      }
      
      // found two siblings, place the new ID between them
      int distance = Math.abs(nid - iid);
      if (distance == 1)
      {
         // no room to move to
         return -1;
      }
      
      return iid + dir * distance / 2;
   }
   
   /**
    * Clone the selected ACL item. 
    */
   private void handleCloneACLItemAs()
   {
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         return;
      }

      Integer cloneId = validateId(cpcdfId, "clone id is invalid");
      if (cloneId == null)
      {
         return;
      }
      
      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);
      
      // busy
      AdminUtils.busy(client);
      
      boolean rc = adm.cloneAcl(row.getResourceType(), 
                                row.getInstance(), 
                                row.getId(), 
                                cloneId);  
      if (!rc)
      {
         // attempt has failed
         client.alarm();

         // idle
         AdminUtils.idle(client);

         return;
      }
      
      // attempt has succeeded; refresh table
      client.ok();
      
      populateACLList();
      
      // locate the clone and make it visible and selected
      sel = locateAcl(row.getResourceType(), row.getInstance(), cloneId);
      if (sel >= 0)
      {
         sel = cpACLs.convertRowIndexToView(sel);
         cpACLsm.setSelectionInterval(sel, sel);
         ensureVisibility(sel);
      }

      handleCancel();

      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Handle Move ACL Up button press - if not first, it will move the ACL
    * one position up.
    */
   private void handleMoveACLUp()
   {
      moveACLItem(true);
   }
   
   /**
    * Handle Move ACL Down button press - if not last, it will move the ACL
    * one position down.
    */
   private void handleMoveACLDown()
   {
      moveACLItem(false);
   }
   
   /**
    * Handle Save button press in direct ID editing dialog.
    * Has the effect of moving the ACL around to an arbitrary position.
    */
   private void handleMoveAclAround()
   {
      // query the selected line
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         return;
      }

      // figure out the source ACL
      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);
      
      Integer targetId = validateId(cpedfId, "target id is invalid");
      if (targetId == null)
      {
         return;
      }
      
      // move it
      moveACL(row, targetId.intValue());
      
      // hide the dialog
      handleCancel();
   }
               
   /**
    * Handle row filtering using a generic filter.
    */
   @SuppressWarnings("unchecked")
   private void handleFilter()
   {
      int    type = JOptionPane.QUESTION_MESSAGE;
      String txt  = "Define the filtering specification below. Press the\n" +
                    "'OK' button without any specification to remove any\n" +
                    "previous filter.";
      String spec = (String) JOptionPane.showInputDialog(client,
                                                         txt,
                                                         "Filter",
                                                         type,
                                                         null,
                                                         null,
                                                         lastFilter);
      if (spec == null || spec.trim().length() == 0)
      {
         // remove filtering
         cpACLsorter.setRowFilter(null);
         lastFilter = null;
      }
      else
      {
         // install a filter
         cpACLsorter.setRowFilter(AdminUtils.createFilter(spec));
         lastFilter = spec;
      }

      String title = screenTitle;
      if (lastFilter != null)
      {
         title += " filtered by <" + lastFilter + ">";
      }
      else
      {
         if (account == null && resourceType == null)
         {
            title += " - unfiltered view";
         }
      }
      
      client.show(SCREEN_ACLS, title);
   }
   
   /**
    * Handle Resequence ACLs button press.
    */
   private void handleResequenceACLs()
   {
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         client.beep("select a row in the target ACL first");
         return;
      }

      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);
      
      // focus on start id
      cpsdfId.requestFocusInWindow();

      // display modal dialog
      dialog = AdminUtils.modalDialog(client, 
                                      "Resequence " + row.getResourceType() + 
                                      " ACLs in " + row.getValueAt(0) +
                                      " instance", 
                                      cpsPanel);
      dialog.setVisible(true);

      // restore state
      if (dialog != null)
      {
         // dialog dismissed by closing window
         handleCancel();
      }
   }
   
   /**
    * Ask the client for a starting number and a step. After this, resequence
    * the ACLs for with the same resource type and instance as the selected 
    * one.  
    */
   private void handleResequenceACLsAs()
   {
      int sel = cpACLs.getSelectedRow();
      
      // safety check
      if (sel == -1)
      {
         return;
      }

      sel = cpACLs.convertRowIndexToModel(sel);
      AclRow row = (AclRow) cpACLtm.data.get(sel);

      Integer startId = validateId(cpsdfId, "starting number is not valid");
      Integer step = validateId(cpsdfStep, "step value is not valid");
      
      if (startId == null || step == null)
      {
         return;
      }
      
      if (startId < 0)
      {
         client.beep("Starting number positive or zero !");
         return;
      }
      
      if (step <= 0)
      {
         client.beep("Step value must be positive !");
         return;
      }
      
      // busy
      AdminUtils.busy(client);
      
      // attempt deletion
      boolean rc = adm.resequenceAcls(row.getResourceType(), 
                                      row.getInstance(), 
                                      startId,
                                      step);
      if (!rc)
      {
         // idle
         AdminUtils.idle(client);
         
         // attempt has failed
         client.alarm();
         return;
      }
      
      // attempt has succeeded; refresh table
      client.ok();
      
      populateACLList();
      cpACLsm.setSelectionInterval(sel, sel);

      handleCancel();
      
      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Handle Cancel button press in any dialog.
    */
   private void handleCancel() 
   {
      // clear error messages
      client.ok();
      
      // hide the dialog
      dialog.setVisible(false);
      dialog = null;
   }
   
   /**
    * Set the state of the screen when no row is selected - some buttons
    * are enabled.
    */
   private void setACLsNoSelection()
   {
      // set enabled state of the list accordingly
      stateMachine(STATE_INITIAL_EMPTY);
   }
   
   /**
    * Set the state of the screen when a single row is selected - all buttons
    * are enabled.
    */
   private void setACLsSingleSelection()
   {
      // set enabled state of the list accordingly
      stateMachine(STATE_ACL_SINGLE);
   }
   
   /**
    * Set the state of the screen when multiple rows are selected - some
    * buttons are enabled.
    */
   private void setACLsMultiSelection()
   {
      // set enabled state of the list accordingly
      stateMachine(STATE_ACL_MULTI);
   }
   
   /**
    * Refresh the ACL list from the server.
    */
   private void handlePopulateACLs()
   {
      // busy
      AdminUtils.busy(client);
      
      doPopulateACLs();
      
      // idle
      AdminUtils.idle(client);
   }
   
   /**
    * Refresh the ACL list from the server.
    */
   private void doPopulateACLs()
   {
      populateACLList();
      setACLsNoSelection();
   }
   
   /**
    * Populate the list of ACLs with fresh data from the server.
    */
   private void populateACLList() 
   {
      // clear the backing list
      acls = null;
      
      // clear selection if any
      cpACLsm.clearSelection();
      
      // clear the table model
      cpACLtm.clear();

      // populate the ACL table
      AclRow[] list = loadAclData();
      
      if (list != null)
      {
         if (list.length > 0)
         {
            cpACLtm.loadAll(list);
         }
      }
      else
      {
         client.beep("no ACLs");
      }
   }
   
   /**
    * Load the ACL items into {@link AclRow} instances and return them. Any
    * menu filtering is applied at this time.
    * <p>
    * If filtering by resource type is in effect, this method takes into
    * account the state of the Effective View check box and the existence of
    * both the shared and private instances. It disables the check box if only
    * one instance is found, but the check box remains selected.
    * 
    * @return   the rows for the data.
    */
   private AclRow[] loadAclData()
   {
      // load the ACLs from server
      if (resourceType != null)
      {
         // filtering by resource type
         // figure out the effective versus full view
         String[] found = adm.listAclInstances(resourceType);
         if (found.length == 0)
         {
            // nothing to show for this resource now
            return null;
         }
         
         int isShar = AdminUtils.isListed("", found);
         int isPriv = AdminUtils.isListed(serverName, found);
         
         if (isShar == -1 && isPriv == -1)
         {
            // nothing to show for the live server
            return null;
         }
         
         if (isShar >= 0 && isPriv >= 0)
         {
            // either effective or full view is possible
            // the views will be different:
            // Effective View check box remains enabled and has effect
            cpcEv.setEnabled(true);
            effectiveView = cpcEv.isSelected();
            
            if (effectiveView)
            {
               // produce the effective view of ACLs
               AclDef aEff = adm.getEffectiveAcls(resourceType);
               acls = new AclDef[] {aEff};
            }
            else
            {
               // produce full view of both shared and private instances
               AclDef aShar = adm.getAcl(resourceType, "");
               AclDef aPriv = adm.getAcl(resourceType, serverName);
               acls = new AclDef[] {aShar, aPriv};
            }
         }
         else
         {
            // effective or full view makes no difference:
            // Effective View check box gets disabled
            cpcEv.setEnabled(false);
            
            String instance = isShar >= 0 ? "" : serverName;
            acls = new AclDef[] {adm.getAcl(resourceType, instance)};
            
            // control the ACL sort order: oid
            effectiveView = true;
         }
      }
      else
      {
         // no filtering by resource type; load full ACLs
         cpcEv.setEnabled(false);
         acls = adm.getAllAcls(null);
         
         // control the ACL sort order: instance, resource type, oid
         effectiveView = false;
      }
      
      if (acls == null)
      {
         return null;
      }
      
      // client-side filtering
      List<AclRow> list = new ArrayList<AclRow>();
      allAcls = new ArrayList<AclRow>();
      
      int errs = 0;
      
      for (int i = 0; i < acls.length; i++)
      {
         errs += acls[i].errors;
         
         for (int j = 0; j < acls[i].acls.length; j++)
         {
            AclDef.Acl acl = acls[i].acls[j];
            
            boolean ok = account == null;
            
            Set<String> subjectsSet = new TreeSet<String>();
            
            StringBuilder subjects = new StringBuilder();
            for (int r = 0; r < acl.subjects.length; r++)
            {
               String acc = acls[i].subjects.get(acl.subjects[r]);
               
               if (acl.subjects[r] == AclDef.ALL_OTHERS)
               {
                  acc = "all_others";
               }
               if (acc == null)
               {
                  acc = "*not found*";
               }

               // filter results by account, if needed
               if (!ok && acc.equals(account))
               {
                  ok = true;
               }

               subjects.append(acc);
               
               if (acl.subjects[r] >= 0)
               {
                  subjectsSet.add(acc);
               }

               if (r < acl.subjects.length - 1)
               {
                  subjects.append(", ");
               }
            }

            AclRow row = new AclRow(acls[i],
                  acl,
                  subjects.toString());
            row.subjectsSet = subjectsSet;

            allAcls.add(row);
            
            // if the specified account is not among the subjects, do not add
            // it to the table.
            if (ok)
            {
               list.add(row);
            }            
         }
      }
      
      // report ACLs in errors
      if (errs > 0)
      {
         String msg = String.format("%d loaded ACLs are with errors", errs);
         client.beep(msg);
      }
      else
      {
         client.ok();
      }
      
      // the entire set is sorted by type and then by id
      Collections.sort(allAcls);
      
      return list.toArray(new AclRow[0]);
   }
   
   /**
    * Validate the given text. Its content are expected to be a valid positive
    * integer value. If the field has no content, it will return
    * <code>null</code>; also, the specified error will be displayed.
    * <p>
    * Else, it will return the field content.
    * 
    * @param    jtf
    *           The source text field
    * @param    err
    *           The error to be displayed if the field has no content.
    * 
    * @return   converted value or <code>null</code> if field is invalid.
    */
   private Integer validateId(JTextField jtf, String err)
   {
      Integer iv = validateIntField(jtf, err);
      
      if (iv == null)
      {
         return null;
      }
      
      // ensure the value is positive
      if (iv.intValue() <= 0)
      {
         // set initial focus on the field
         jtf.requestFocusInWindow();
         
         // alarm
         client.beep(err);
         return null;
      }
      
      return iv;
   }
   
   /**
    * Switches the ACL view according to the new resource type.
    * The reason for this method to exist is a centralized Effective View
    * mode control.
    * 
    * @param    newType
    *           resource type to switch to or <code>null</code> for unfiltered
    *           view
    */
   private void switchResourceType(String newType)
   {
      // prevent unwanted handler invocation
      changingEffective = true;
      
      // if a different resource view, enable and set Effective View
      if (newType != null && !newType.equals(resourceType))
      {
         cpcEv.setEnabled(true);
         cpcEv.setSelected(true);
      }
      
      // if no filtering by resource type, disable and reset Effective View
      else if (newType == null)
      {
         cpcEv.setEnabled(false);
         cpcEv.setSelected(false);
      }
         
      // prevent unwanted handler invocation
      changingEffective = false;
      
      // otherwise, staying with the same resource type, no change
      resourceType = newType;
   }
   
   /**
    * Looks up the list of all ACLs and locates the one with given resource
    * type, instance and ID.
    * 
    * @param    resourceType
    *           resource type to search within
    * 
    * @param    instanceName
    *           ACL instance name to search within
    * 
    * @param    id
    *           ACL id of interest
    * 
    * @return   index of the requested ACL in the list or <code>-1</code>
    */
   private int locateAcl(String resourceType, String instanceName, int id)
   {
      for (int i = 0; i < allAcls.size(); i++)
      {
         AclRow row = allAcls.get(i);
         if (!instanceName.equals(row.getInstance()))
         {
            continue;
         }
         if (!resourceType.equals(row.getResourceType()))
         {
            continue;
         }
         if (id == row.getId())
         {
            return i;
         }
      }
      
      return -1;
   }
   
   /**
    * Makes sure the specified row in the ACL table is in the viewport.
    * 
    * @param    row
    *           table row index (view, not model)
    */
   private void ensureVisibility(int row)
   {
      Rectangle rect = cpACLs.getCellRect(row, 0, true);
      cpACLs.scrollRectToVisible(rect);
   }
   
   /**
    * A class holding all necessary info related to accounts.
    */
   private static class AccountRow 
   extends TaggedName
   {
      /** Type field index. */
      private static final int IDX_TYPE = IDX_LAST + 1;
      
      /**
       * Constructor.
       * 
       * @param    type
       *           The type of this account.
       * @param    account
       *           The account.
       */
      AccountRow(String type, TaggedName account)
      {
         super(new String[] {account.getName(), account.getTag(), type});
      }

      /**
       * Compares this instance of the class with another.
       * 
       * @param    o
       *           Object to compare with.
       * 
       * @return   the usual integers required by Comparable
       */
      public int compareTo(TaggedName o)
      {
         AccountRow row = (AccountRow) o;
         
         if (getType().equals(row.getType()))
         {
            return getName().compareTo(row.getName());
         }

         return getType().compareTo(row.getType());
      }

      /**
       * Get ACL type.
       * 
       * @return   the type
       */
      private String getType()
      {
         return get(IDX_TYPE);
      }
   }
   
   /**
    * A class holding all necessary info related to ACL items, to be 
    * represented in a table.
    */
   private class AclRow 
   extends TaggedName
   {
      /** The string representation for the subjects list.*/
      private String subjects;

      /** A set with all subject names.*/
      private Set<String> subjectsSet;
      
      /** The ACL item for this row. */
      private AclDef.Acl acl;
      
      /** The ACL parent definition for this row. */
      private AclDef aclDef;

      /**
       * Constructor.
       * 
       * @param   aclDef
       *          The ACL parent definition for this row.
       * @param   acl
       *          The acl item.
       * @param   subjects
       *          The string representation for the subjects list.
       */
      AclRow(AclDef     aclDef,
             AclDef.Acl acl,
             String     subjects)
      {
         super((String) null, null);
         this.aclDef      = aclDef;
         this.acl         = acl;
         this.subjects    = subjects;
         this.subjectsSet = Collections.emptySet();
      }
                     
      /**
       * Compares this instance of the class with another. The native ordering
       * is as follows:
       * <ul>
       * <li>by the instance name
       * <li>within instance by the resource type name
       * <li>within instance and resource type by the object ID
       * </ul>
       * In the effective view mode, the native order no longer applies. Since
       * the effective view is only possible for a single resource type, the
       * effective ordering is just the object ID ordering.
       * 
       * @param    o
       *           Object to compare with.
       * 
       * @return   the usual integers required by Comparable
       */
      public int compareTo(TaggedName o)
      {
         AclRow row = (AclRow) o;
         
         if (!effectiveView)
         {
            if (!getInstance().equals(row.getInstance()))
            {
               return getInstance().compareTo(row.getInstance());
            }
         }
         
         if (!getResourceType().equals(row.getResourceType()))
         {
            return getResourceType().compareTo(row.getResourceType());
         }
         
         return acl.oid - ((AclRow) o).acl.oid;
      }

      /**
       * Get a value for a certain column.
       * 
       * @param    col
       *           The column for which the value is needed.
       * 
       * @return   The content of this column.
       */
      public Object getValueAt(int col)
      {
         switch (col)
         {
            case 0 :
            {
               String instance = getInstance();
               return instance.length() == 0 ? SHARED_INSTANCE
                                            : instance;
            }
            case 1 :
            {
               return getResourceType();
            }
            case 2 :
            {
               return String.format("% 10d", getId());
            }
            case 3 :
            {
               return getName();
            }
            case 4 :
            {
               return getSubjects();
            }
            case 5 :
            {
               return getRights();
            }
         }

         return null;
      }
   
      /**
       * Compares this instance of the class with another. Comparison is
       * case-insensitive.
       * 
       * @param    o
       *           Object to compare with.
       * 
       * @return   <code>true</code> when both are equal
       */
      public boolean equals(Object o)
      {
         if (o == null || !(o instanceof AclRow))
         {
            return false;
         }
         
         AclRow row = (AclRow) o;
         return getResourceType().equals(row.getResourceType()) && 
                getInstance().equals(row.getInstance())         &&
                getId() == row.getId();
                
      }

      /**
       * Get the instance name for this item.
       * 
       * @return   instance name for this item.
       */
      private String getInstance()
      {
         if (acl.origin != null)
         {
            return acl.origin;
         }
         
         return aclDef.server;
      }
      
      /**
       * Get the resource type for this item.
       * 
       * @return   the resource type for this item.
       */
      private String getResourceType()
      {
         return aclDef.resourceName;
      }
      
      /**
       * Get the id for this item.
       * 
       * @return   the id for this item.
       */
      private int getId()
      {
         return acl.oid;
      }
      
      /**
       * Get the Resource Instance Name for this item.
       * 
       * @return   Resource Instance Name for this item.
       */
      public String getName()
      {
         return acl.name;
      }
      
      /**
       * Set the Resource Instance Name for this item.
       * 
       * @param    name
       *           The name to be set.
       */
      public void setName(String name)
      {
         acl.name = name;
      }
      
      /**
       * Get the exact nature of this item.
       * 
       * @return   <code>true</code> if exact; <code>false</code> if regexp.
       */
      private boolean getExact()
      {
         return acl.exact;
      }
      
      /**
       * Set the exact nature of this item
       * 
       * @param    exact
       *           <code>true</code> if exact; <code>false</code> if regexp.
       */
      private void setExact(boolean exact)
      {
         acl.exact = exact;
      }
      
      /**
       * Get a string representation of the subjects.
       * 
       * @return   string representation of the subjects
       */
      private String getSubjects()
      {
         return subjects;
      }
      
      /**
       * Get a string representation of the rights.
       * 
       * @return   string representation of the rights.
       */
      private String getRights()
      {
         if (acl.rights == null)
         {
            return "*** error ***";
         }
         
         return acl.rights.toString();
      }
      
      /**
       * Get the OID error flag.
       * 
       * @return   <code>true</code> if OID is invalid
       */
      private boolean isErrorOid()
      {
         return acl.errOid;
      }
      
      /**
       * Get the resource instance name error flag.
       * 
       * @return   <code>true</code> if the resource instance name is invalid
       */
      private boolean isErrorName()
      {
         return acl.errName;
      }
      
      /**
       * Get the resource instance name nature error flag.
       * 
       * @return   <code>true</code> if the resource instance name nature is
       *           invalid
       */
      private boolean isErrorExact()
      {
         return acl.errExact;
      }
      
      /**
       * Get the subjects error flag.
       * 
       * @return   <code>true</code> if subjects is invalid
       */
      private boolean isErrorSubjs()
      {
         return acl.errSubjs;
      }
      
      /**
       * Get the rights error flag.
       * 
       * @return   <code>true</code> if rights is invalid
       */
      private boolean isErrorRights()
      {
         return acl.errRights;
      }
      
      /**
       * Clear the OID error flag.
       */
      private void clearErrorOid()
      {
         acl.errOid = false;
      }
      
      /**
       * Clear the resource instance name error flag.
       */
      private void clearErrorName()
      {
         acl.errName = false;
      }

      /**
       * Clear the resource instance name nature error flag.
       */
      private void clearErrorExact()
      {
         acl.errExact = false;
      }

      /**
       * Clear the subjects error flag.
       */
      private void clearErrorSubjs()
      {
         acl.errSubjs = false;
      }

      /**
       * Clear the rights error flag.
       */
      private void clearErrorRights()
      {
         acl.errRights = false;
      }
   }

   /**
    * A table model for the ACL screen.
    */
   private class AclTableModel extends TaggedDataTableModel
   {
      /** Default column sizes. */
      private final int columnSizes[] =
      {
         100,
         100,
         80,
         150,
         250,
         90,
      };
      
      /**
       * C'tor.
       */
      public AclTableModel()
      {
         super("Instance", 
               "Resource Type",
               "ID",
               "Resource Instance Name",
               "Subjects",
               "Rights"); 
      }

      /**
       * Returns the column class for this model.
       * 
       * @param    col
       *           model column index
       * 
       * @return   column class, which is always <code>String.class</code>
       */
      public Class<?> getColumnClass(int col)
      {
         return String.class;
      }
   
      /**
       * Gets the width of the given column.
       * 
       * @param    col
       *           The column index (0-based).
       * 
       * @return   The width in pixels.
       */
      public int getColumnWidth(int col)
      {
         return columnSizes[col];
      }
   }
   
   /**
    * Custom renderer for the ACL table.
    */
   private class AclTableRenderer extends DefaultTableCellRenderer
   {
      private Font EXACT_FONT = null;
      
      /**
       * Visualize ACL errors. Set the font to bold for the resource item  
       * name column, when the nature is <b>exact</b>.
       */
      @Override
      public Component getTableCellRendererComponent(JTable table,
                                                     Object value, 
                                                     boolean isSelected, 
                                                     boolean hasFocus, 
                                                     int row,
                                                     int column)
      {
         Component renderer = super.getTableCellRendererComponent(
               table, value, isSelected, hasFocus, row, column);

         if (value == null || row < 0)
         {
            return renderer;
         }
         
         // get the underlying ACL
         int drow = cpACLs.convertRowIndexToModel(row);
         AclRow acl = (AclRow) cpACLtm.data.get(drow);
         
         // customize renderer according to error status and options
         if (!isSelected)
         {
            renderer.setBackground(null);
         }
         
         // oid column
         if (column == 2 && acl.acl.errOid)
         {
            renderer.setBackground(Color.RED);
         }
         
         // resource instance name column 
         if (column == 3)
         {
            if (acl.acl.errName || acl.acl.errExact)
            {
               renderer.setBackground(Color.RED);
            }
            else if (acl.acl.exact)
            {
               if (EXACT_FONT == null)
               {
                  EXACT_FONT = renderer.getFont().deriveFont(Font.BOLD);
                  
               }
               renderer.setFont(EXACT_FONT);
            }
         }
         
         // subjects column
         if (column == 4)
         {
            if (acl.acl.errSubjs)
            {
               renderer.setBackground(Color.RED);
            }
            else
            {
               // split the subject text with 10 subjects per line
               StringBuilder tooltip = new StringBuilder("<html>");
               int idx = 0;
   
               Iterator<String> itr = acl.subjectsSet.iterator();
               while (itr.hasNext())
               {
                  tooltip.append(itr.next());
                  if (itr.hasNext())
                  {
                     tooltip.append(", ");
                  }
   
                  idx++;
                  if (idx % 7 == 0)
                  {
                     tooltip.append("<br/>");
                  }
               }
   
               tooltip.append("</html>");
               this.setToolTipText(tooltip.toString());
            }
         }

         // rights column
         if (column == 5 && acl.acl.errRights)
         {
            renderer.setBackground(Color.RED);
         }
         
         return /* this;*/ renderer;
      }
   }
}
